﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2016 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/twi.h>
#include "AVRChip.h"
#include "Timer.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define DSPL_SET B,2                                // -+
#define PIR      C,1                                //  |
#define DCF      C,2                                //  |
#define SSR      C,3                                //  |
#define P1       D,0                                //  | Aufgabenzuordnung der Port-Pins
#define P2       D,1                                //  |
#define ENC_A    D,2                                //  |
#define ENC_B    D,3                                //  |
#define ENC_PB   D,4                                //  |
#define SYS_LED  D,7                                // -+

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                            // Ereignisse der DCF-PWM Abtastung:
{   DCF_0    = '0',                     // · Null erkannt
    DCF_1    = '1',                     // · Eins erkannt
    DCF_MARK = 'm',                     // · Minutenanfang erkannt
    DCF_FAIL = 'e',                     // · Abtastfehler
    DCF_NONE = 'x'                      // · Keine Ereignisse
} DCFEvent;

typedef struct                          // Tonabfolge:
{   const Tone* tone;                   // · Ton
    UInt08      index;                  // · Tonfolge-Index
    UInt08      hold;                   // · Resthaltedauer [10 ms]
} ToneSequence;

//////////////////////////////////////////////////////////////////////////////////////////
//  Konstanten  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static const SInt08 I2C_ADR_DS1307 = 0b1101000;     // I²C Geräteadresse der RTC

static const UInt08  toneFreqRegMap[][2] PROGMEM =  // Tonfrequenz zu Registerwerte
{   //  N ,  OCR       Index  Enum       Ton   = Frequenz [Hz] | µC-Freq [Hz] (Abweichung)
    //---------------------------------------------------------+--------------------------
    { 0x00, 0x00 }, // [ 0]   NOTE_XX    Stille                |
    { 0x40, 0xDB }, // [ 1]   NOTE_C3    C3    =    130.813    |    130.909     ( 0.07%)
    { 0x40, 0xCF }, // [ 2]   NOTE_C3f   C3#   =    138.591    |    138.462     (-0.09%)
    { 0x40, 0xC3 }, // [ 3]   NOTE_D3    D3    =    146.832    |    146.939     ( 0.07%)
    { 0x40, 0xB8 }, // [ 4]   NOTE_D3f   D3#   =    155.563    |    155.676     ( 0.07%)
    { 0x40, 0xAE }, // [ 5]   NOTE_E3    E3    =    164.814    |    164.571     (-0.15%)
    { 0x40, 0xA4 }, // [ 6]   NOTE_F3    F3    =    174.614    |    174.545     (-0.04%)
    { 0x40, 0x9B }, // [ 7]   NOTE_F3f   F3#   =    184.997    |    184.615     (-0.21%)
    { 0x40, 0x92 }, // [ 8]   NOTE_G3    G3    =    195.998    |    195.918     (-0.04%)
    { 0x40, 0x8A }, // [ 9]   NOTE_G3f   G3#   =    207.652    |    207.194     (-0.22%)
    { 0x40, 0x82 }, // [10]   NOTE_A3    A3    =    220.000    |    219.847     (-0.07%)
    { 0x40, 0x7B }, // [11]   NOTE_B3    B3    =    233.082    |    232.258     (-0.35%)
    { 0x40, 0x74 }, // [12]   NOTE_H3    H3    =    246.942    |    246.154     (-0.32%)
    { 0x40, 0x6D }, // [13]   NOTE_C4    C4    =    261.626    |    261.818     ( 0.07%)
    { 0x40, 0x67 }, // [14]   NOTE_C4f   C4#   =    277.183    |    276.923     (-0.09%)
    { 0x40, 0x61 }, // [15]   NOTE_D4    D4    =    293.665    |    293.878     ( 0.07%)
    { 0x40, 0x5C }, // [16]   NOTE_D4f   D4#   =    311.127    |    309.677     (-0.47%)
    { 0x40, 0x56 }, // [17]   NOTE_E4    E4    =    329.628    |    331.034     ( 0.43%)
    { 0x40, 0x51 }, // [18]   NOTE_F4    F4    =    349.228    |    351.220     ( 0.57%)
    { 0x40, 0x4D }, // [19]   NOTE_F4f   F4#   =    369.994    |    369.231     (-0.21%)
    { 0x40, 0x48 }, // [20]   NOTE_G4    G4    =    391.995    |    394.521     ( 0.64%)
    { 0x40, 0x44 }, // [21]   NOTE_G4f   G4#   =    415.305    |    417.391     ( 0.50%)
    { 0x40, 0x40 }, // [22]   NOTE_A4    A4    =    440.000    |    443.077     ( 0.70%)
    { 0x40, 0x3D }, // [23]   NOTE_B4    B4    =    466.164    |    464.516     (-0.35%)
    { 0x40, 0x39 }, // [24]   NOTE_H4    H4    =    493.883    |    496.552     ( 0.54%)
    { 0x40, 0x36 }, // [25]   NOTE_C5    C5    =    523.251    |    523.636     ( 0.07%)
    { 0x40, 0x33 }, // [26]   NOTE_C5f   C5#   =    554.365    |    553.846     (-0.09%)
    { 0x40, 0x30 }, // [27]   NOTE_D5    D5    =    587.330    |    587.755     ( 0.07%)
    { 0x40, 0x2D }, // [28]   NOTE_D5f   D5#   =    622.254    |    626.087     ( 0.62%)
    { 0x40, 0x2B }, // [29]   NOTE_E5    E5    =    659.255    |    654.545     (-0.71%)
    { 0x40, 0x28 }, // [30]   NOTE_F5    F5    =    698.456    |    702.439     ( 0.57%)
    { 0x40, 0x26 }, // [31]   NOTE_F5f   F5#   =    739.989    |    738.462     (-0.21%)
    { 0x40, 0x24 }, // [32]   NOTE_G5    G5    =    783.991    |    778.378     (-0.72%)
    { 0x40, 0x22 }, // [33]   NOTE_G5f   G5#   =    830.609    |    822.857     (-0.93%)
    { 0x40, 0x20 }, // [34]   NOTE_A5    A5    =    880.000    |    872.727     (-0.83%)
    { 0x08, 0xF6 }, // [35]   NOTE_B5    B5    =    932.328    |    932.794     ( 0.05%)
    { 0x08, 0xE8 }, // [36]   NOTE_H5    H5    =    987.767    |    988.841     ( 0.11%)
    { 0x08, 0xDB }, // [37]   NOTE_C6    C6    =   1046.502    |   1047.273     ( 0.07%)
    { 0x08, 0xCF }, // [38]   NOTE_C6f   C6#   =   1108.731    |   1107.692     (-0.09%)
    { 0x08, 0xC3 }, // [39]   NOTE_D6    D6    =   1174.659    |   1175.510     ( 0.07%)
    { 0x08, 0xB8 }, // [40]   NOTE_D6f   D6#   =   1244.508    |   1245.405     ( 0.07%)
    { 0x08, 0xAE }, // [41]   NOTE_E6    E6    =   1318.510    |   1316.571     (-0.15%)
    { 0x08, 0xA4 }, // [42]   NOTE_F6    F6    =   1396.913    |   1396.364     (-0.04%)
    { 0x08, 0x9B }, // [43]   NOTE_F6f   F6#   =   1479.978    |   1476.923     (-0.21%)
    { 0x08, 0x92 }, // [44]   NOTE_G6    G6    =   1567.982    |   1567.347     (-0.04%)
    { 0x08, 0x8A }, // [45]   NOTE_G6f   G6#   =   1661.219    |   1657.554     (-0.22%)
    { 0x08, 0x82 }, // [46]   NOTE_A6    A6    =   1760.000    |   1758.779     (-0.07%)
    { 0x08, 0x7B }, // [47]   NOTE_B6    B6    =   1864.655    |   1858.065     (-0.35%)
    { 0x08, 0x74 }, // [48]   NOTE_H6    H6    =   1975.533    |   1969.231     (-0.32%)
    { 0x08, 0x6D }, // [49]   NOTE_C7    C7    =   2093.005    |   2094.545     ( 0.07%)
    { 0x08, 0x67 }, // [50]   NOTE_C7f   C7#   =   2217.461    |   2215.385     (-0.09%)
    { 0x08, 0x61 }, // [51]   NOTE_D7    D7    =   2349.318    |   2351.020     ( 0.07%)
    { 0x08, 0x5C }, // [52]   NOTE_D7f   D7#   =   2489.016    |   2477.419     (-0.47%)
    { 0x08, 0x56 }, // [53]   NOTE_E7    E7    =   2637.020    |   2648.276     ( 0.43%)
    { 0x08, 0x51 }, // [54]   NOTE_F7    F7    =   2793.826    |   2809.756     ( 0.57%)
    { 0x08, 0x4D }, // [55]   NOTE_F7f   F7#   =   2959.955    |   2953.846     (-0.21%)
    { 0x08, 0x48 }, // [56]   NOTE_G7    G7    =   3135.963    |   3156.164     ( 0.64%)
    { 0x08, 0x44 }, // [57]   NOTE_G7f   G7#   =   3322.438    |   3339.130     ( 0.50%)
    { 0x08, 0x40 }, // [58]   NOTE_A7    A7    =   3520.000    |   3544.615     ( 0.70%)
    { 0x08, 0x3D }, // [59]   NOTE_B7    B7    =   3729.310    |   3716.129     (-0.35%)
    { 0x08, 0x39 }, // [60]   NOTE_H7    H7    =   3951.066    |   3972.414     ( 0.54%)
    { 0x08, 0x36 }, // [61]   NOTE_C8    C8    =   4186.009    |   4189.091     ( 0.07%)
    { 0x08, 0x33 }, // [62]   NOTE_C8f   C8#   =   4434.922    |   4430.769     (-0.09%)
    { 0x08, 0x30 }, // [63]   NOTE_D8    D8    =   4698.636    |   4702.041     ( 0.07%)
    { 0x08, 0x2D }, // [64]   NOTE_D8f   D8#   =   4978.032    |   5008.696     ( 0.62%)
    { 0x08, 0x2B }, // [65]   NOTE_E8    E8    =   5274.041    |   5236.364     (-0.71%)
    { 0x08, 0x28 }, // [66]   NOTE_F8    F8    =   5587.652    |   5619.512     ( 0.57%)
    { 0x08, 0x26 }, // [67]   NOTE_F8f   F8#   =   5919.911    |   5907.692     (-0.21%)
    { 0x08, 0x24 }, // [68]   NOTE_G8    G8    =   6271.927    |   6227.027     (-0.72%)
    { 0x08, 0x22 }, // [69]   NOTE_G8f   G8#   =   6644.875    |   6582.857     (-0.93%)
    { 0x08, 0x20 }, // [70]   NOTE_A8    A8    =   7040.000    |   6981.818     (-0.83%)
    { 0x01, 0xF6 }, // [71]   NOTE_B8    B8    =   7458.620    |   7462.348     ( 0.05%)
    { 0x01, 0xE8 }, // [72]   NOTE_H8    H8    =   7902.133    |   7910.730     ( 0.11%)
    { 0x01, 0xDB }, // [73]   NOTE_C9    C9    =   8372.018    |   8378.182     ( 0.07%)
    { 0xFF, 0x00 }  // [74]   Marker für "Tonsequenz wiederholen"
};

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16        msecTicks       = 0;
static volatile UInt16        secTicks        = 0;
static volatile DCFEvent      dcfEvent        = DCF_NONE;
static volatile bool          keyPressEvent   = false;
static volatile UInt08        keyPressTime    = 0;
static volatile SInt08        encoderPosition = 0;
static volatile ToneSequence  toneSequence    = { NULL, 0, 0 };
static volatile DateTime      sysDateTime     = DATETIME_INVALID;
static volatile UInt08        sys10ms         = 0;
static          bool          doMainLoop      = true;
static          SInt08        ambBrightness   = 0;
static          SInt08        ledBrightness   = 0;
static          DCFStatus     dcfStatus       = { 0, 60000 };
static          bool          rtcOk           = false;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool
ReadI2C(SInt08 adr, UInt08* buffer, UInt08 num)
{
    bool result = false;
    // I²C Einheit einschalten und Start-Kennung senden
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
    UInt08 t;
    for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us(); // max. 1ms warten
    if( t > 0 )
    {   if( TW_STATUS == TW_START )
        {   // Zielgeräte-Addresse mit Leseanforderung übertragen (SLA+R)
            TWDR = ((UInt08)adr << 1) | TW_READ;
            TWCR = BV(TWINT) | BV(TWEN);
            for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us();
            if( t > 0 )
            {   if( TW_STATUS == TW_MR_SLA_ACK )
                {   // <num> Datenbytes entgegennehmen, das letzte unbestätigt (NACK)
                    UInt08 i;
                    for( i=0; i<num; i++ )
                    {   UInt08  acknowledge;
                        if( i < num-1 )
                        {   TWCR = BV(TWINT) | BV(TWEN) | BV(TWEA);
                            acknowledge = TW_MR_DATA_ACK;
                        }
                        else
                        {   TWCR = BV(TWINT) | BV(TWEN);
                            acknowledge = TW_MR_DATA_NACK;
                        }
                        for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us();
                        if( t == 0 || TW_STATUS != acknowledge ) break;
                        buffer[i] = TWDR;
                    }
                    result = i == num;
                }
            }
        }
    }
    // Stop-Kennung senden und I²C Einheit abschalten
    TWCR = BV(TWINT)|BV(TWSTO)|BV(TWEN);
    Freeze100us();
    TWCR = 0;
    return result;
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
WriteI2C(SInt08 adr, const UInt08* buffer, UInt08 num)
{
    bool result = false;
    // I²C Einheit einschalten und Start-Kennung senden
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
    UInt08 t;
    for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us(); // max. 1ms warten
    if( t > 0 )
    {   if( TW_STATUS == TW_START )
        {   // Zielgeräte-Addresse mit Schreib-Anforderung senden (SLA+W)
            TWDR = ((UInt08)adr << 1) | TW_WRITE;
            TWCR = BV(TWINT)|BV(TWEN);
            for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us();
            if( t > 0 )
            {   if( TW_STATUS == TW_MT_SLA_ACK )
                {   // <num> Datenbytes übertragen
                    UInt08 i;
                    for( i=0; i<num; i++ )
                    {   TWDR = buffer[i];
                        TWCR = BV(TWINT)|BV(TWEN);
                        for( t=100; !(TWCR & BV(TWINT)) && t>0; t-- ) Freeze10us();
                        if( t == 0 || TW_STATUS != TW_MT_DATA_ACK ) break;
                    }
                    result = i == num;
                }
            }
        }
    }
    // Stop-Kennung senden und I²C Einheit abschalten
    TWCR = BV(TWINT)|BV(TWSTO)|BV(TWEN);
    Freeze100us();
    TWCR = 0;
    return result;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DS1307_Read(): Liest die Echtzeituhr vom Typ DS1307 aus. Schlägt das Auslesen fehl, //
//  -------------  kehrt die Funktion mit 'false' zurück. Datum und Zeit der 'dateTime' //
//  sind in dem Fall ungültig. Kehrt die Funktion mit 'true' zurück, aber Datum und     //
//  Zeit in 'dateTime' sind dennoch ungültig, steht die Echtzeituhr.                    //
//                                                                                      //
//  Eingabe: · dateTime: Zeiger auf die Datumzeit-Struktur, die Datum und Zeit auf-     //
//                       nimmt (darf nicht NULL sein).                                  //
//  Ausgabe: 'true', wenn die Echtzeituhr ausgelesen werden konnte.						//
//////////////////////////////////////////////////////////////////////////////////////////
static bool
DS1307_Read(DateTime* dateTime)
{
    UInt08  data[7];
    DateTime invalid = DATETIME_INVALID;
    *dateTime = invalid;
    data[0] = 0;
    rtcOk = false;
    if( !WriteI2C(I2C_ADR_DS1307, data, 1) ) return false;
    if( !ReadI2C(I2C_ADR_DS1307, data, CARDINALITY(data)) ) return false;
    rtcOk = true;
    if( data[0] & BV(7) ) return true;                                // "clock halt"?
    dateTime->time.second =        (data[0]>>4)*10 + (data[0]&0x0F);  // Sekunde
    dateTime->time.minute =        (data[1]>>4)*10 + (data[1]&0x0F);  // Minute
    dateTime->time.hour   =        (data[2]>>4)*10 + (data[2]&0x0F);  // Stunde
    dateTime->date.day    =        (data[4]>>4)*10 + (data[4]&0x0F);  // Tag
    dateTime->date.month  =        (data[5]>>4)*10 + (data[5]&0x0F);  // Monat
    dateTime->date.year   = 2000 + (data[6]>>4)*10 + (data[6]&0x0F);  // Jahr
    return DATETIME_IsValid(dateTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DS1307_Write(): Stellt die Hardwareuhr vom Typ DS1307 und setzt sie gegebenenfalls  //
//  --------------  in Gang. Ist die übergebene Zeit&Datumsinformation ungültig, wird   //
//  die Hardwareuhr angehalten.                                                         //
//                                                                                      //
//  Eingabe: · dateTime: Zeiger auf die Datumzeit-Struktur, die Datum und Zeit bereit-  //
//                       bereitstellt (darf nicht NULL sein).                           //
//  Ausgabe: 'true', wenn der Zugriff auf die Echtzeituhr erfolgreich war.              //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
DS1307_Write(const DateTime* dateTime)
{
    UInt08  data[9];

    data[0] = 0;
    if( DATETIME_IsValid(dateTime) )
    {   data[1] = (dateTime->time.second/10<<4)+(dateTime->time.second%10);
        data[2] = (dateTime->time.minute/10<<4)+(dateTime->time.minute%10);
        data[3] = (dateTime->time.hour/10<<4)+(dateTime->time.hour%10); // 24h-Darstellung
        data[4] = DATE_WeekDay(&dateTime->date);
        data[5] = (dateTime->date.day/10<<4)+(dateTime->date.day%10);
        data[6] = ((dateTime->date.month)/10<<4)+((dateTime->date.month)%10);
        data[7] = ((dateTime->date.year%100)/10<<4)+(dateTime->date.year%10);
        data[8] = 0x10;       // Ausgabe 1Hz Rechecksignal an Pin 7 des DS1307 einschalten
        rtcOk = WriteI2C(I2C_ADR_DS1307, data, CARDINALITY(data));
    }
    else
    {   data[1] = 0x80;
        rtcOk = WriteI2C(I2C_ADR_DS1307, data, 2);  // Uhr (nur) anhalten
    }
    return rtcOk;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
WriteSystemClock(const DateTime* dateTime)
{
    if( DATETIME_IsValid(dateTime) )
    {   Atomic
        {   sysDateTime = *dateTime;
            sys10ms     = 0;
        }
    }
    else
    {   Atomic
        {   sysDateTime.time.second = 255;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
ReadADC(UInt08 channel) // Kanalnummer: 0..7
{
    UInt08  hi;
    UInt08  lo;

//  ADMUX =             channel;  // Eingangsspannung bezogen auf VRef
    ADMUX = BV(REFS0) | channel;  // Eingangsspannung bezogen auf AVcc
    ADCSRA |=  BV(ADSC);
    loop_until_bit_is_clear(ADCSRA, ADSC);
    lo  = ADCL;
    hi  = ADCH;
    return ((UInt16)(hi & 0x03) << 8) | lo;   // Ausgabe: 0..1023
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedDcfDecoding(void)
{
    static const UInt08  FAIL    = 0xFF;    // Dekodierungsfehler (255 nicht ändern!)
    static const UInt08  BCD[]   = { 1, 2, 4, 8, 10, 20, 40, 80 };
    static bool          dstTime = false;
    static UInt08        minute  = 0;      // 0 .. 59
    static UInt08        hour    = 0;      // 0 .. 23
    static UInt08        day     = 1;      // 1 .. 31
    static UInt08        month   = 1;      // 1 .. 12
    static UInt08        year    = 0;      // (2000+) 0 .. 99

    // DCF-Funkausfalldauer inkrementieren
    static Timer timer = TIMER(TIMER_SECONDS, 60);
    if( TIMER_RestartIfExpired(&timer) )
    {   if( dcfStatus.elapsed < 60000 ) dcfStatus.elapsed++;
    }
    // Ereignis aus der PWM-Auswertung holen
    DCFEvent  event;
    Atomic
    {   event    = dcfEvent;
        dcfEvent = DCF_NONE;
    }
    // Aktuellen Bitzähler = Aktuelle Sekunde holen
    UInt08 bitNo = dcfStatus.bitCount;
    if( event == DCF_MARK && bitNo == 59 )
    {   // Neue Zeit!
        DateTime  dateTime;
        dateTime.time.second = 0;
        dateTime.time.minute = minute;
        dateTime.time.hour   = hour;
        dateTime.date.day    = day;
        dateTime.date.month  = month;
        dateTime.date.year   = 2000 + year;
        if( DATETIME_IsValid(&dateTime) )
        {	SInt08  shiftHrs;
            if( dstTime ) shiftHrs = -2;    // MESZ -> UTC
            else          shiftHrs = -1;    // MEZ  -> UTC
            DATETIME_ShiftHour(&dateTime, shiftHrs);
            WriteSystemClock(&dateTime);    // Empf. Zeit in die Systemuhr übertragen
            DS1307_Write(&dateTime);        // Empf. Zeit in die Hardwareuhr übertragen
            TIMER_Restart(&timer);
            dcfStatus.elapsed = 0;
        }
    }
    // 0-1-Zustand oder Fehler aus der PWM-Auswertung lesen
    bool dcfBit = false;
    switch( event )
    {   case DCF_NONE: return;
        case DCF_0:    dcfBit = false; break;
        case DCF_1:    dcfBit = true;  break;
        default:       bitNo  = FAIL;
    }
    // Dekodierung der Zeitinformation fortsetzen
    static bool parity = false;
    parity ^= dcfBit;
    switch( bitNo )
    {   case  0:                                              // -+
        case  1:                                              //  |
        case  2:                                              //  |
        case  3:                                              //  |
        case  4:                                              //  |
        case  5:                                              //  |
        case  6:                                              //  |
        case  7:                                              //  | Bits 0 - 16
        case  8:                                              //  | ignorieren
        case  9:                                              //  |
        case 10:                                              //  |
        case 11:                                              //  |
        case 12:                                              //  |
        case 13:                                              //  |
        case 14:                                              //  |
        case 15:                                              //  |
        case 16:                                       break; // -+
        case 17: dstTime = dcfBit;                     break; // MESZ
        case 18: if( dcfBit == dstTime ) bitNo = FAIL; break; // MEZ
        case 19:                                       break; // Schaltsekunde
        case 20: if( !dcfBit ) bitNo = FAIL;           break; // Startbit (1)
        case 21: parity = dcfBit; minute = 0;                 // -+
        case 22:                                              //  |
        case 23:                                              //  | Minute
        case 24:                                              //  |
        case 25:                                              //  |
        case 26:                                              //  |
        case 27: if( dcfBit ) minute += BCD[bitNo-21]; break; // -+
        case 28: if( parity ) bitNo = FAIL;            break; // Parität Minute
        case 29: parity = dcfBit; hour = 0;                   // -+
        case 30:                                              //  |
        case 31:                                              //  | Stunde
        case 32:                                              //  |
        case 33:                                              //  |
        case 34: if( dcfBit ) hour += BCD[bitNo-29];   break; // -+
        case 35: if( parity ) bitNo = FAIL;            break; // Parität Stunde
        case 36: parity = dcfBit; day  = 0;                   // -+
        case 37:                                              //  | Kalender-
        case 38:                                              //  | tag:
        case 39:                                              //  | 1 .. 31
        case 40:                                              //  |
        case 41: if( dcfBit ) day += BCD[bitNo-36];    break; // -+
        case 42:                                              // -+ Wochentag
        case 43:                                              //  |  ignorieren
        case 44:                                       break; // -+
        case 45: month = 0;                                   // -+
        case 46:                                              //  | Monat:
        case 47:                                              //  |  1 .. 12
        case 48:                                              //  |
        case 49: if( dcfBit ) month += BCD[bitNo-45];  break; // -+
        case 50: year = 0;                                    // -+
        case 51:                                              //  |
        case 52:                                              //  |
        case 53:                                              //  |  Jahr:
        case 54:                                              //  |  0 .. 99
        case 55:                                              //  |
        case 56:                                              //  |
        case 57: if( dcfBit ) year += BCD[bitNo-50];   break; // -+
        case 58: if( parity ) bitNo = FAIL;            break; // Parität Datum
        default: bitNo = FAIL;
    }
    bitNo++;                    // bitNo: 0 .. 58 oder 255 (=FAIL) -> 0 .. 59
    dcfStatus.bitCount = bitNo; // dcfStatus.bitCount: 0 .. 59
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSysClockSyncing(void)
{
    static Timer  update = TIMER(TIMER_SECONDS, CFG_SYS_CLOCK_SYNC_INTERVAL*60);

    if( TIMER_RestartIfExpired(&update) )
    {   if( BOARD_DCFStatus()->elapsed > 13 )
        {   // Letzter Funkzeitempfang liegt mehr als 13 Minuten zurück.
            DateTime dateTime;
            DS1307_Read(&dateTime);                    // Hardwareuhr auslesen und
			if( rtcOk ) WriteSystemClock(&dateTime);   // Systemuhr damit stellen
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedAmbientBrightness(void)
{
    const  SInt16  min    =   5; // Grenzwert für   0%,  0    .. max-1
    const  SInt16  max    = 200; // Grenzwert für 100%, min+1 ..  255
    static Timer   update = TIMER(TIMER_MILLISECONDS, CFG_AMB_BRIGHT_UPDATE);

    if( !TIMER_RestartIfExpired(&update) ) return;
    SInt16  value;
    // ADC auslesen und in % umwandeln
    value = 1023 - ReadADC(0);  // Inveretieren (LDR ist gegen GND geschaltet)
    value >>= 2;                // [0 .. 1023] -> [0 .. 255]
    if     ( value < min ) ambBrightness =    0;
    else if( value > max ) ambBrightness =  100;
    else                   ambBrightness = (value-min)*100 / (max-min); // %
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();    // Interrupts sperren
    wdt_enable(WDTO_15MS);  // Watchdog-Intervall auf kürzeste Zeit setzen
    while( true );          // Auf "Zubiss" des Watchdogs warten
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER2_COMPA_vect)
{
    // Abweichungskompensation              // Quarztaktbedingt findet der IRQ-Aufruf
    {   static UInt08 counter = 0;          // bei einem Zeitgeber-Zählerwert von 115
        if( counter == 4 )                  // nicht exakt jede Millisekunde statt,
        {   counter = 0;                    // sondern etwas schneller:
            OCR2A = F_CPU/(32*1000UL);      //
        }                                   //		115 ÷ 115.2 = 0.9983 ms
        else                                //
        {   counter++;                      // Zwecks Ausgleichs dieser Ungenauigkeit
            OCR2A = F_CPU/(32*1000UL)-1;    // wird bei jedem fünften Aufruf, der
        }                                   // Überlauf um einen Zählerwert verzögert:
    }    									// 4·(115÷115.2) + (116÷115.2) = 5 ms
    // Zeitsignale
    struct
    {   bool    event10ms  : 1;     // Tritt im   10 ms Intervall ein
        bool    event100ms : 1;     // Tritt im  100 ms Intervall ein
        bool    eventSec   : 1;     // Tritt im 1000 ms Intervall ein
    } timeSignal = { false, false, false };
    {   static UInt08 counter1ms   = 0;
        static UInt08 counter10ms  = 0;
        static UInt08 counter100ms = 0;
        counter1ms++;
        timeSignal.event10ms = counter1ms == 10;
        if( counter1ms == 10 )
        {   counter1ms  = 0;
            counter10ms++;
            timeSignal.event100ms = counter10ms == 10;
            if( counter10ms == 10 )
            {   counter10ms  = 0;
                counter100ms++;
                timeSignal.eventSec = counter100ms == 10;
                if( counter100ms == 10 ) counter100ms = 0;
            }
        }
    }
    // Millisekundenticker inkrementieren
    msecTicks++;
    // Sekundenticker inkrementieren
    if( timeSignal.eventSec ) secTicks++;
    // Systemzeit aktualisieren
    if( timeSignal.event10ms )
    {   sys10ms++;
        if( sys10ms == 100 )
        {   sys10ms =  0;
            if( sysDateTime.time.second < 60 )
            {   DATETIME_ShiftSecond((DateTime*)&sysDateTime, 1);
            }
        }
    }
    // Behandlung Drehinkrementalgeber (dazu s. artikel.grasbon.de/Drehinkrementalgeber)
    //////////////////////////////////////////////////////////////////////////////////////
    //                                                                                  //
    //      A-B-Signaländerungen bei Bewegung des Drehinkrementalgebers:                //
    //                                                                                  //
    //                               _______         _______         _                  //
    //               A:  ··· _______| :   : |_______| :   : |_______| ···               //
    //                       ___  :   :  _______  :   :  _______  :                     //
    //               B:  ···    |_______| :   : |_______| :   : |_____···               //
    //                            :   :   :   :   :   :   :   :   :                     //
    //                            :   :   :   :   :   :   :   :   :                     //
    //                           -4  -3  -2  -1   0  +1  +2  +3  +4                     //
    //                                                                                  //
    //                                                                                  //
    //  Abfrage mittels Mealy-Automaten M = { S, I, O, zt, s0, F }                      //
    //                                                                                  //
    //    S  = { S0, .. ,S3 }                                                           //
    //    I  = { 00,01,10,11 }                                                          //
    //    O  = { left,none,right }      Zustandsdiagramm:                               //
    //    zt : S×I -> S×O                                                               //
    //    s0 = S0                         00,right         +--------+ 10,right          //
    //    F  = Ø                             +------------>|////////+------------+      //
    //                                       |    01,left  |// S0 //|            |      //
    //                                       |   +---------+////////|<-------+   |      //
    //   Eingaben I:                         |   |         +--------+        |   |      //
    //    · 00: Signalpaar A/B = 0/0         |   |                           |   |      //
    //    · 01: Signalpaar A/B = 0/1         |  \|/                  00,left |  \|/     //
    //    · 10: Signalpaar A/B = 1/0       +-+------+                     +--+-----+    //
    //    · 11: Signalpaar A/B = 1/1       |////////|                     |////////|    //
    //                                     |// S1 //|                     |// S2 //|    //
    //                                     |////////|                     |////////|    //
    //                                     +-----+--+                     +------+-+    //
    //   Ausgaben O:                        /|\  | 11,left                  /|\  |      //
    //    · left:  Drehungbewegung           |   |                           |   |      //
    //             nach links                |   |         +--------+        |   |      //
    //    · right: Drehungbewegung           |   +-------->|////////+--------+   |      //
    //             nach rechts               |             |// S3 //| 10,left    |      //
    //    · none:  Keine Bewegung            +-------------+////////|<-----------+      //
    //                                            01,right +--------+        11,right   //
    //                                                                                  //
    //   Zustandsübergangsfunktion zt:                                                  //
    //                                                                                  //
    //      zt |    00    |    01    |    10    |    11    |                            //
    //     ----+----------+----------+----------+----------+                            //
    //      S0 | S0,none  | S1,left  | S2,right | S3,none  |                            //
    //      S1 | S0,right | S1,none  | S2,none  | S3,left  |                            //
    //      S2 | S0,left  | S1,none  | S2,none  | S3,right |                            //
    //      S3 | S0,none  | S1,right | S2,left  | S3,none  |                            //
    //     ----+----------+----------+----------+----------+                            //
    //                                                                                  //
    //                                                                                  //
    //////////////////////////////////////////////////////////////////////////////////////
    {
        #if CFG_ROT_ENC_TYPE == 1
        // Inkrementalgeber: DDM Hopt+Schuler --------------------------------------------
        static const SInt08 encZt[][4] PROGMEM = { {  0, -1, +1,  0 },
                                                   { +1,  0,  0, -1 },
                                                   { -1,  0,  0, +1 },
                                                   {  0, +1, -1,  0 }
                                                 };
        static const SInt08 rasterStop = 2;
        #elif CFG_ROT_ENC_TYPE == 2
        // Inkrementalgeber: Alps --------------------------------------------------------
        static const SInt08 encZt[][4] PROGMEM = { {  0,  0, +1,  0 },
                                                   {  0,  0,  0, -1 },
                                                   { -1,  0,  0,  0 },
                                                   {  0, +1,  0,  0 }
                                                 };
        static const SInt08 rasterStop = 2;
        #elif CFG_ROT_ENC_TYPE == 3
        // Inkrementalgeber: EN11 von AB / TT --------------------------------------------
        static const SInt08 encZt[][4] PROGMEM = {  { 0,  0,  0,  0 },
                                                    { 0,  0,  0,  0 },
                                                    { 0,  0,  0,  0 },
                                                    { 0, -1, +1,  0 }
                                                 };
        static const SInt08 rasterStop = 1;
        // -------------------------------------------------------------------------------
        #else
         #error Undefined encoder type in CFG_ROT_ENC_TYPE
        #endif
        static UInt08  encState = 0;
        static SInt08  encDelta = 0;
        UInt08 encAB = (!GetPin(ENC_A) << 1) + !GetPin(ENC_B);
        encDelta += (SInt08)pgm_read_byte(&(encZt[encState][encAB]));
        encState = encAB;
        UInt08 position = encoderPosition;
        if( encDelta == rasterStop )
        {   if( position <  127 ) position++;
            encDelta = 0;
        }
        else if( encDelta == -rasterStop )
        {   if( position > -127 ) position--;
            encDelta = 0;
        }
        encoderPosition = position;
    }
    // Tastenkontrolle: Drehinkrementalgeber-Taster
    {   static bool pressedState = false;
        if( timeSignal.event10ms )
        {   static UInt08 latency   = 0;
            bool          isPressed = !GetPin(ENC_PB);
            if( pressedState )
            {   // Taster geschlossen
                if( !isPressed && latency > 0 ) latency--;
                else                            latency = CFG_ENC_UP_RESPONSIVITY;
            }
            else
            {   // Taster geöffnet
                if( isPressed && latency > 0 ) latency--;
                else                           latency = CFG_ENC_DOWN_RESPONSIVITY;
                keyPressTime = 0;
            }
            if( latency == 0 )
            {   pressedState  = isPressed;
                keyPressEvent = isPressed;
            }
        }
        if( (timeSignal.event100ms) && pressedState && keyPressTime < 250 )
        {   keyPressTime++;
        }
    }
    // Steuerung Tonfolge
    if( timeSignal.event10ms )
    {   if( toneSequence.hold > 0 )
        {   toneSequence.hold--;
            if( toneSequence.hold == 0 )
            {   UInt08 index = toneSequence.index;
                toneSequence.index++;
                Tone  tone;
                tone.pitch   = pgm_read_byte(&(toneSequence.tone[index].pitch));
                tone.sustain = pgm_read_byte(&(toneSequence.tone[index].sustain));
                TCCR1B = 0;
                TCNT1  = 0;
                UInt08 n = pgm_read_byte(&toneFreqRegMap[tone.pitch][0]);
                OCR1A    = pgm_read_byte(&toneFreqRegMap[tone.pitch][1]);
                switch( n )
                {   case 0x01:
                        TCCR1A = BV(COM1A0);
                        TCCR1B = BV(WGM12) | BV(CS10);
                        break;
                    case 0x08:
                        TCCR1A = BV(COM1A0);
                        TCCR1B = BV(WGM12) | BV(CS11);
                        break;
                    case 0x40:
                        TCCR1A = BV(COM1A0);
                        TCCR1B = BV(WGM12) | BV(CS11) | BV(CS10);
                        break;
                    case 255:  // Tonabfolge wiederholen
                        toneSequence.index = 0;
                    default:    // Tonabfolge beenden
                        TCCR1A = 0;
                        TCCR1B = 0;
                }
                toneSequence.hold = tone.sustain;
            }
        }
    }
    // Abtastung des DCF77 PWM-Signals (dazu. s. Zeitschrift c't Hacks, 2.2014, S. 84)
    //////////////////////////////////////////////////////////////////////////////////////
    //                                                                                  //
    //  Mealy-Automat zur DCF77 PWM Abtastung:                                          //
    //  --------------------------------------                                          //
    //                                                                                  //
    //                                                                                  //
    //   Darstellung '0':    Abtastung '0':                                             //
    //                                                                                  //
    //     |                 Hi ··+-----------+    :      :      :      :      :        //
    //   1·+---+                  |//////:////|    :      :      :      :      :        //
    //     |///|              S0  |/ S1 /: S2 | S3 :  S4  :  S5  :  S6  :  S11 :        //
    //     |///|                  |//////:////|    :      :      :      :      :        //
    //   0·+---+---+--->     Lo ··+//////:////+---------------------------------···     //
    //        100     ms     -----+-----------------------+------------------------>    //
    //                            :      :         :      :      :      :      :   t    //
    //                            0      T0        T1     T2     T3     T4     T5       //
    //                                                                                  //
    //   Darstellung '1':    Abtastung '1':                                             //
    //                                                                                  //
    //     |                 Hi ··+-------------------------+    :      :      :        //
    //   1·+-------+              |//////://////://////:////|    :      :      :        //
    //     |///////|          S0  |/ S1 /:/ S2 /:/ S7 /: S8 | S9 :  S10 :  S11 :        //
    //     |///////|              |//////://////://////:////|    :      :      :        //
    //   0·+---+---+--->     Lo ··+//////://////://////:////+-------------------···     //
    //            200 ms     -----+--------------------+--------------------------->    //
    //                            :      :      :      :         :      :      :    t   //
    //                            0      T0     T1     T2        T3     T4     T5       //
    //                                                                                  //
    //                                                                                  //
    //   Mealy-Automat M:                                                               //
    //                                                                                  //
    //   M = { S,I,O,zt,s0,F }                             +------+                     //
    //                                                +--->|//////|                     //
    //     S  = { S0, .. ,S12 }                  LO,x |    |/ S0 /|                     //
    //     I  = { TI,HI,LO }                          +----+//////|                     //
    //     O  = { x,a,0,1,m,e }                            +----+-+        LO,x         //
    //     zt : S×I -> S×O                                      |          +---+        //
    //     s0 = S0                                     +-----+  | HI,a     |   |        //
    //     F  = Ø                                 HI,x |    \|/\|/         |  \|/       //
    //                                                 |   +------+      +-+-----+      //
    //                                                 +---+//////| HI,m |///////|      //
    //   Eingaben I:                                       |/ S1 /|<-----+/ S12 /|<--+  //
    //     TI = Zeitpunkt T0..T6 erreicht  +-------------->|//////|      |///////|   |  //
    //     HI = High-Signal liegt vor    __|               +--+---+      +-------+   |  //
    //     LO = Low-Signal              |     LO,x            |             HI,x     |  //
    //                                  |     +--+            |             +--+     |  //
    //     Beispielzeiten für TI:       |     |  |       TI,x |             |  |     |  //
    //       T0 =   60 ms               |     | \|/          \|/           \|/ |     |  //
    //       T1 =  150 ms               |   +-+----+       +------+       +----+-+   |  //
    //       T2 =  170 ms               |   |//////|  LO,x |//////| TI,x  |//////|   |  //
    //       T3 =  250 ms               |   |/ S3 /|<------+/ S2 /+------>|/ S7 /|   |  //
    //       T4 =  900 ms               |   |//////|       |//////|       |//////|   |  //
    //       T5 = 1200 ms               |   +--+---+       +----+-+       +---+--+   |  //
    //       T6 = 2200 ms               |      |            /|\ |             |      |  //
    //                                  |      | TI,x        |  |        TI,x |      |  //
    //   Ausgaben O:                    |     \|/            +--+            \|/     |  //
    //     x = Kein Ereignis            |   +------+         HI,x         +------+   |  //
    //     a = Beginn                   |   |//////|<---+            +--->|//////|   |  //
    //     0 = '0' erkannt              |   |/ S4 /|    | LO,x  HI,x |    |/ S8 /|   |  //
    //     1 = '1' erkannt              |   |//////+----+            +----+//////|   |  //
    //     m = Minutenanfang erkannt    |   +--+---+                      +---+--+   |  //
    //     e = Abtastfehler             |      |                              |      |  //
    //                                  |      | TI,x                    LO,x |      |  //
    //                                  |     \|/                            \|/     |  //
    //   Zustandsübergangsfunktion zt:  |   +------+                      +------+   |  //
    //                                  |   |//////|<---+            +--->|//////|   |  //
    //        |  TI  |  HI  |  LO  |    |   |/ S5 /|    | LO,x  LO,x |    |/ S9 /|   |  //
    //    ----+------+------+------+    |   |//////+----+            +----+//////|   |  //
    //     S0 | S0,e | S1,a | S0,x |    |   +--+---+        LO,x          +---+--+   |  //
    //     S1 | S2,x | S1,x | S0,e |    |      |            +---+             |      |  //
    //     S2 | S7,x | S2,x | S3,x |    |      | TI,x       |   |        TI,x |      |  //
    //     S3 | S4,x | S0,e | S3,x |    |     \|/           |  \|/           \|/     |  //
    //     S4 | S5,x | S0,e | S4,x |    |   +------+      +-+-----+       +-------+  |  //
    //     S5 | S6,x | S0,e | S5,x |    |   |//////| TI,0 |///////|  TI,1 |///////|  |  //
    //     S6 |S11,0 | S0,e | S6,x |    |   |/ S6 /+----->|/ S11 /|<------+/ S10 /|  |  //
    //     S7 | S8,x | S7,x | S0,e |    |   |//////|      |///////|       |///////|  |  //
    //     S8 | S0,e | S8,x | S9,x |    |   +-+----+      +-+---+-+       +-+-----+  |  //
    //     S9 |S10,x | S0,e | S9,x |    |     | /|\         |   |           | /|\    |  //
    //    S10 |S11,1 | S0,e |S10,x |    |     |  | LO,x     |   |      LO,x |  |     |  //
    //    S11 |S12,x | S1,a |S11,x |    |     +--+          |   |           +--+     |  //
    //    S12 | S0,e | S1,m |S12,x |    |              HI,a |   | TI,x               |  //
    //    ----+------+------+------+    +-------------------+   +--------------------+  //
    //                                                                                  //
    //////////////////////////////////////////////////////////////////////////////////////
    if( timeSignal.event10ms )
    {
        typedef enum
        {   TI = 0,
            HI = 1,
            LO = 2
        } Input;

        typedef struct
        {   UInt08  state;
            char    output;
        } Entry;

        static const Entry zeta[][3] PROGMEM =
        {   //+--Ti--+ +--Hi--+ +--Lo--+ +--------+
            { { 0,'e'},{ 1,'a'},{ 0,'x'} }, // S0   //  Eingaben:
            { { 2,'x'},{ 1,'x'},{ 0,'e'} }, // S1   //   TI = Zeitpunkt i erreicht
            { { 7,'x'},{ 2,'x'},{ 3,'x'} }, // S2   //   HI = DCF-High-Signal liegt vor
            { { 4,'x'},{ 0,'e'},{ 3,'x'} }, // S3   //   LO = DCF-Low-Signal  liegt vor
            { { 5,'x'},{ 0,'e'},{ 4,'x'} }, // S4
            { { 6,'x'},{ 0,'e'},{ 5,'x'} }, // S5   //  Ausgaben:
            { {11,'0'},{ 0,'e'},{ 6,'x'} }, // S6   //   x = Kein Ereignis
            { { 8,'x'},{ 7,'x'},{ 0,'e'} }, // S7   //   a = Beginn
            { { 0,'e'},{ 8,'x'},{ 9,'x'} }, // S8   //   0 = '0' erkannt
            { {10,'x'},{ 0,'e'},{ 9,'x'} }, // S9   //   1 = '1' erkannt
            { {11,'1'},{ 0,'e'},{10,'x'} }, // S10  //   m = Minutenanfang erkannt
            { {12,'x'},{ 1,'a'},{11,'x'} }, // S11  //   e = Abtastfehler
            { { 0,'e'},{ 1,'m'},{12,'x'} }  // S12
        };

        static UInt08  state = 0;
        static UInt08  time  = 0;
        Input          input;
        char           output;

        if( time ==    6 ||              // T0
            time ==   15 ||              // T1
            time ==   17 ||              // T2
            time ==   25 ||              // T3
            time ==   95 ||              // T4
            time ==  120 ||              // T5
            time ==  220  ) input = TI;  // T6
        else if( GetPin(DCF)    ) input = HI;
        else                      input = LO;
        output = pgm_read_byte(&(zeta[state][input].output));
        state  = pgm_read_byte(&(zeta[state][input].state));
        if( output == 'm' ) time = 0;
        if     ( output == 'a' ) time     = 0;
        else if( output != 'x' ) dcfEvent = (DCFEvent)output;
        time++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife in Main. Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller     //
//  betroffenen Module ermöglichen einen kontrollierten Abschluss der Aktivitäten.      //
//  Die letzte aufgerufenen Funktion ist die "Shutdown"-Funktion dieses Board-Moduls,   //
//  die dann den physikalischer Reset auslöst (s. BOARD_Shutdown() bzw. Reset() ).      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doMainLoop = false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Liest 'size' Bytes ab Adresse 'adr' aus dem EEPROM-Speicher des    //
//  ---------------  Controllers und legt sie in das Bytefeld ab 'mem' ab.              //
//                                                                                      //
//  Eingabe: · adr:  Adresse im EEPROM-Speicher.                                        //
//           · mem:  Adresse des Byteblocks.                                            //
//           · size: Anzahl der zu lesenden Bytes.                                      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* adr, void* mem, UInt16 size)
{
    eeprom_read_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Legt den Inhalt des durch 'mem' und 'size' spezifizierten Byte-    //
//  ---------------  blocks in den EEPROM-Speicher ab Adresse 'adr' ab. Da Lösch- und   //
//  Schreibvorgänge auf EEPROM-Speicher nur begrenzt oft möglich sind (lt. Datenblatt   //
//  ca. 100000 Write/Erase-Cycles) wird vor dem Speichern geprüft, ob der zu schrei-    //
//  bende Wert sich vom Inhalt der Speicherzelle unterscheidet und ob sich der Schreib- //
//  vorgang an der betroffenen Speicherstelle erübrigt.                                 //
//  Hinweis: Man beachte, dass das Beschreiben einer EEPROM-Speicherzelle ca. 8 ms in   //
//           Anspruch nimmt und damit verhältnismäßig viel Zeit verschlingt.            //
//                                                                                      //
//  Eingabe: · adr:  Adresse im EEPROM-Speicher.                                        //
//           · mem:  Adresse des Byteblocks.                                            //
//           · size: Größe des Byteblocks in Bytes.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(void* adr, const void* mem, UInt16 size)
{
    eeprom_update_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SendDisplay(): Überträgt das Byte in 'data' in das Register 'reg' des LED-    //
//  -------------------  Siebensegmenttreibers vom Typ MAX7221.                         //
//                                                                                      //
//  Eingabe: · reg:  Nummer des zu beschreibenden MAX7221-Registers: 0 .. 15.           //
//           · data: Der Registerwert: 0 .. 255.                                        //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SendDisplay(UInt08 reg, UInt08 data)
{
    ClrPin(DSPL_SET);
    SPDR = reg;
    loop_until_bit_is_set(SPSR, SPIF);
    SPDR = data;
    loop_until_bit_is_set(SPSR, SPIF);
    SetPin(DSPL_SET);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_StartToneSequence(): Initiiert das interruptgesteuerte Abspielen einer Ton-   //
//  -------------------------  folge. Das in 'sequence' angegebene Feld von 'Tone'-     //
//  Objekten muss als abschließenden Eintrag TONE_SEQ_END (Tonfolge beenden) oder       //
//  TONE_SEQ_REPEAT (Tonfolge wiederholen) haben.                                       //
//                                                                                      //
//  Eingabe: · sequence: Adresse auf das abzuspielende Feld von 'Tone'-Objekten.        //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_StartToneSequence(const Tone* sequence)
{
    Atomic
    {   toneSequence.tone  = sequence;
        toneSequence.index = 0;
        toneSequence.hold  = 1;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_StopToneSequence(): Beendet unmittelbar die gerade laufende Tonfolge.         //
//  ------------------------                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_StopToneSequence(void)
{
    static const Tone mute PROGMEM = TONE_SEQ_END;

    BOARD_StartToneSequence(&mute);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Clock(): Liest die Systemuhr aus. In dem Fall, dass die Systemuhr nicht läuft //
//  -------------  ist die zurückgelieferte Datumzeit ungültig. Die Systemuhr liefert   //
//  die Zeit und das Datum in UTC.                                                      //
//                                                                                      //
//  Eingabe: · dateTime: Referenz, die die Zeitinformation aufnimmt.                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Clock(DateTime* dateTime)
{
    Atomic
    {   *dateTime = sysDateTime;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetClock(): Überträgt die in 'dateTime' angegebene Datumzeit in die System-   //
//  ----------------  und die Hardwareuhr. Ist die übergebene Datumzeit ungültig,       //
//  werden die Uhren angehalten. Die Zeitinformation wird als UTC erwartet.             //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, mit der die Systemuhr gestellt wird.                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetClock(const DateTime* dateTime)
{
    WriteSystemClock(dateTime);
    DS1307_Write(dateTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_DCFStatus(): Liefert den Zustand der Funkzeit-Empfangseinheit. Die Rückgabe   //
//  -----------------  ist ein Zeiger auf eine Struktur vom Typ 'DCFStatus'.            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Zeiger auf die Informations-Struktur vom Typ 'DCFStatus'.                  //
//////////////////////////////////////////////////////////////////////////////////////////
const DCFStatus*
BOARD_DCFStatus(void)
{
    return &dcfStatus;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_RTClockFailure(): Gibt es Kommunikationsproblem mit der externen Echtzeituhr, //
//  ----------------------  kehrt diese Funktion mit 'true' zurück. Die Echtzeituhr     //
//  wird vom System in ProceedSysClockSyncing() ausgelesen bzw. in ProceedDcfDecoding() //
//  gestellt.                                                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Kommunikationsproblem mit der RTC auftrat.				//
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_RTClockFailure(void)
{
    return !rtcOk;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_AmbientBrightness(): Liefert die aktuelle Umgebungshelligkeit in Prozent.     //
//  -------------------------                                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Umgebungshelligkeit: 0 .. 100 [%].                                         //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
BOARD_AmbientBrightness(void)
{
    return ambBrightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetLedBrightness(): Bestimmt die Helligkeit der Bedienteil-LED.               //
//  ------------------------                                                            //
//                                                                                      //
//  Eingabe: · brightness: Helligkeit der LED: 0 .. 100 [%].                            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLedBrightness(SInt08 brightness)
{
    SInt16  dutyCycle;

    if     ( brightness > 100 ) brightness = 100;
    else if( brightness <   0 ) brightness =   0;
    dutyCycle = ((SInt16)brightness*(SInt16)brightness)/39;   // Gammakorrektur 0.5
    if( dutyCycle > 255 ) OCR0A = 255;
    else                  OCR0A = (UInt08)dutyCycle;
    ledBrightness = brightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_LedBrightness(): Liefert die Helligkeit der Bedienteil-LED.                   //
//  ---------------------                                                               //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Helligkeit der LED: 0 .. 100 [%].                                          //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
BOARD_LedBrightness(void)
{
    return ledBrightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PIRSensor(): Gibt den Wert des passiven Infrarot-Bewegungsmelders zurück.     //
//  -----------------                                                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sensor Bewegung detektiert hat.							//
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_PIRSensor(void)
{
    return GetPin(PIR);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_TakeKey(): Liefert 'true', wenn seit der letzten Abfrage mit BOARD_TakeKey()  //
//  ---------------  die Taste gedrückt und gehalten wurde.                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Tastendruck-Ereignis stattgefunden hat.                   //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_TakeKey(void)
{
    bool  ret = false;

    Atomic
    {   ret = keyPressEvent;
        keyPressEvent = false;
    }
    return ret;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_KeyTime(): Liefert die Dauer des Tastendrucks. Ist die Taste aktuell nicht    //
//  ---------------  gedrückt, kehrt die Funktion mit 0 zurück. Ist die Taste länger    //
//  als 25 Sekunden gedrückt, bleibt der Zeitzähler auf 25 s (= 250 s÷10) stehen.       //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Zeit seit dem letzten Tastendruck-Ereignis bis jetzt: 1 .. 250 [s÷10].     //
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
BOARD_KeyTime(void)
{
    UInt08 value = 0;

    Atomic
    {   value = keyPressTime;
    }
    return value;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_TakeEncoder(): Liefert die Veränderung des Drehinkrementgebers seit der       //
//  -------------------  letzten Abfrage mit BOARD_TakeEncoder(). Positive Werte be-    //
//  deuten eine Drehbewegung um entsprechend viele Rasterpositionen im Uhrzeigersinn,   //
//  negative Wert entsprechend umgekehrt.                                               //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Veränderung der Drehposition: -127 .. +127 [Rasterpositionen].             //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
BOARD_TakeEncoder(void)
{
    SInt08 value = 0;

    Atomic
    {   value = encoderPosition;
        encoderPosition = 0;
    }
    return value;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnableRelais(): Ändert den Zustand des Solid-State-Relais.                    //
//  --------------------                                                                //
//                                                                                      //
//  Eingabe: · enable: 'true'/'false' = Relais ein-/ausschalten.                        //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnableRelais(bool enable)
{
    ChgPin(SSR, enable);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnableP1(): Ändert den Zustand des P1-Pins.									//
//  ----------------																	//
//                                                                                      //
//  Eingabe: · enable: 'true'/'false' = P1 ein-/aus.									//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnableP1(bool enable)
{
    ChgPin(P1, enable);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnableP2(): Ändert den Zustand des P2-Pins.									//
//  ----------------																	//
//                                                                                      //
//  Eingabe: · enable: 'true'/'false' = P2 ein-/aus.									//
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnableP2(bool enable)
{
    ChgPin(P2, enable);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  System  //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(B, _L_,    // PB.0: ---     : ---
                _L_,    // PB.1: [OC1A]  : Piezosignalgeber, Frequenz über Timer 1 PWM
                _H_,    // PB.2: DSPL_SET: MAX7219 LOAD (Anzeige)
                _L_,    // PB.3: [MOSI]  : MAX7219 DIN  (Anzeige)
                _L_,    // PB.4: ---     : ---
                _L_,    // PB.5: [SCK]   : MAX7219 CLK  (Anzeige)
                _T_,    // PB.6: ---     : Quarz
                _T_);   // PB.7: ---     : Quarz
    ConfPort(C, _T_,    // PC.0: [ADC0]  : Helligkeitssensor, 0 = Hell
                _T_,    // PC.1: PIR     : PIR-Bewegungssensor, H = Bewegung
                _T_,    // PC.2: DCF     : DCF-Modul, H = Signal
                _L_,    // PC.3: SSR     : Solid State Relais LED, H = ein
                _T_,    // PC.4: [SDA]   : I²C
                _T_,    // PC.5: [SCL]   : I²C
                _T_,    // PC.6: ---     : Reset
                _T_);   //
    ConfPort(D, _L_,    // PD.0: P1      : H = Feiertag oder Wochenende
                _L_,    // PD.1: P2      : H = Weckalarm
                _T_,    // PD.2: ENC_A   : Drehinkrementalgeber Signal A
                _T_,    // PD.3  ENC_B   : Drehinkrementalgeber Signal B
                _T_,    // PD.4: ENC_PB  : Drehinkrementalgeber Taste, L = gedrückt
                _L_,    // PD.5: ---     : ---
                _L_,    // PD.6: [OC0A]  : Konsolen LED, Helligkeit über Timer 0 PWM
                _L_);   // PD.7: SYS_LED : System-Led auf der Platine, H = ein
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // ADCs ------------------------------------------------------------------------------
    ADCSRA  = BV(ADEN) | BV(ADPS2) | (ADPS1) | (ADPS0);  // ADC einschalten (F_CPU÷128)
    ADCSRA |= ADSC;                                      // 1. Abtastung sofort anstoßen
    // Timer/Counter 0 als Phase Correct PWM an OC0A mit (F_CPU÷32)÷256÷2 = 225Hz --------
    TCCR0A |= BV(WGM00) | BV(COM0A1);
    TCCR0B |= BV(CS01)  | BV(CS00);
    OCR0A  = 0;
    // Timer/Counter 1 als CTC an OC1A --- (Frequenzgenerator für den Signalgeber) -------
    TCCR1A |= BV(COM1A0);               // freq = F_CPU÷(2·N·(1+OCR1A))
    TCCR1B |= BV(WGM12);                //        (N = Vorteiler (1,8,64,256,1024))
    OCR1A  = 0;
    // Timer 2: Clear Timer on Compare Match (CTC) Mode, Intervall 0.9983 ms -------------
    TCCR2A = BV(WGM21);
    TCCR2B = BV(CS21) | BV(CS20);
    OCR2A  = F_CPU/(32*1000UL)-1;   // 32 = Vorteilung, 1000 = Wiederholrate
    TIMSK2 = BV(OCIE2A);
    // SPI Master, F_CPU÷2, MSB first, Sample on rising Edge, Mode 0 ---------------------
    SPCR = BV(SPE)| BV(MSTR);
    SPSR = BV(SPI2X);
    // I²C Master, SCL freq = F_CPU ÷ (16 + 2·TWBR · 4^TWPS) = 17 kHz --------------------
    TWCR = BV(TWEN);
    TWAR = 0;
    TWBR = 100;     // TWBR = 100, Bitratenstellgröße
    TWSR = 0;       // TWPS = 0  , Keine Vorteilung
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doMainLoop;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    ProceedDcfDecoding();
    ProceedSysClockSyncing();
    ProceedAmbientBrightness();
    wdt_reset();
    UInt16 ticks = BOARD_MSecTicks();
    ChgPin(SYS_LED, (bool)((ticks&0x0380)==0x0080 || (ticks&0x0380)==0x0180));
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    Reset();
}

//////////////////////////////////////////////////////////////////////////////////////////
