﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Menü                                                                                //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2016 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "Sound.h"
#include "Storing.h"
#include "Display.h"
#include "String.h"
#include "Timer.h"
#include "Menu.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                // Menüpunkte
{   SHOW_ALARMTIME = 0,     // · Zeige aktuelle Weckzeit
    ENABLE_ALARM   = 1,     // · Wecker einschalten
    EXIT           = 2      // · Menü verlassen
} MenuItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Timer     exitTimer      = TIMER(TIMER_SECONDS, CFG_MODE_MENU_TIMEOUT);
static MenuItem  menuItem       = SHOW_ALARMTIME;
static bool      setupAlarmTime = false;
static Time      alarmTime      = DEF_ALARM_TIME;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  PrintTime(): Druckt die angegeben Zeit in den String. Die Ausgabe umfasst Stunden   //
//  -----------  und Minuten, wobei bei den Stunden eine führende Null durch ein Leer-  //
//  zeichen ersetzt wird. Ist die Zeit ungültig, erhält der String "-- --".             //
//                                                                                      //
//  Eingabe: · string: String, der die Zeitausgabe aufnimmt.                            //
//           · time:   Die auszugebende Zeit.                                           //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintTime(String* string, const Time* time)
{
    if( time->hour < 24 && time->minute < 60 )
    {   if( time->hour < 10 ) STRING_PrintChar(string, ' ');
        STRING_PrintNum(string, time->hour);
        STRING_PrintChar(string, '-');
        if( time->minute < 10 ) STRING_PrintChar(string, '0');
        STRING_PrintNum(string, time->minute);
    }
    else
    {   STRING_PrintLit(string, "-- --");
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ShiftTime(): Verschiebt die in 'time' angegebene Zeit um 'shift' Schritte. Die      //
//  -----------  Schrittweite ist dabei 5 Minuten. Bei Stundenwechseln wird die Zeit    //
//  auf volle Stunden gesetzt wird. Zu jedem Zeitverschiebungsschritt wird ein Tick-    //
//  Geräusch generiert.                                                                 //
//                                                                                      //
//  Eingabe: · time:  Die auszugebende Zeit.                                            //
//           · shift: Verschiebung der Zeit: -127 .. +128.                              //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ShiftTime(Time* time, SInt08 shift)
{
    while( shift > 0 )
    {   SOUND_Start(SOUND_ENC_TICK);
        time->minute += 5;
        if( time->minute > 55 )
        {   time->minute = 0;
            time->hour++;
            if( time->hour > 23 ) time->hour = 0;
        }
        shift--;
    }
    while( shift < 0 )
    {   SOUND_Start(SOUND_ENC_TICK);
        if( time->minute < 5 )
        {   if( time->hour == 0 ) time->hour = 24;
            time->hour--;
            time->minute = 60;
        }
        time->minute -= 5;
        shift++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ShiftMenuItem(): Versetzt den in 'item' angegebenen Menupunkt um 'shift' Einträge.  //
//  ---------------  Ist 'item' an den Menü-Grenzen angekommen, bleibt die Funktion     //
//  ohne Wirkung. Zu jedem Menupunktwechsel wird ein Tick-Geräusch generiert.           //
//                                                                                      //
//  Eingabe: · item:  Der zu versetzende Menupunkt.                                     //
//           · shift: Anzahl der Wechsel: -127 .. +128.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ShiftMenuItem(MenuItem* item, SInt08 shift)
{
    while( shift > 0 )
    {   if( *item != EXIT )
        {   *item  = (MenuItem)((UInt08)(*item)+1);
            SOUND_Start(SOUND_ENC_TICK);
        }
        shift--;
    }
    while( shift < 0 )
    {   if( *item != SHOW_ALARMTIME )
        {   *item  = (MenuItem)((UInt08)(*item)-1);
            SOUND_Start(SOUND_ENC_TICK);
        }
        shift++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
MENU_Enter(void)
{
    menuItem = SHOW_ALARMTIME;
    DISPLAY_SetBlinkingDigits(0);
    DISPLAY_StartAnim(ANIM_MENU_OPEN);
    SOUND_Start(SOUND_OPEN_MENU);
	TIMER_Restart(&exitTimer);
    STORING_Load(STO_AlarmTime, &alarmTime);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
MENU_Proceed(void)
{
    // Ausgabe
    switch( menuItem )
    {   case SHOW_ALARMTIME:
        {   String string = STRING;
            STRING_PrintLit(&string, "AL ");
            PrintTime(&string, &alarmTime);
            DISPLAY_Print(&string);
            break;
        }
        case ENABLE_ALARM:
            if( ALARM_IsEnabled() ) DISPLAY_PrintLit("AL  AUS ");
            else                    DISPLAY_PrintLit("AL  Ein ");
            break;
        default:
            DISPLAY_PrintLit("AL  --- ");
    }
    // Aktionen
    if( APPLICATION_Key() )
    {   if( setupAlarmTime )
        {   // Weckzeit stellen beenden
			alarmTime.second = 0;
            STORING_Save(STO_AlarmTime, &alarmTime);
            ALARM_Enable(true);
            SOUND_Start(SOUND_ALARM_ENABLE);
            DISPLAY_StartAnim(ANIM_ALARM_ENABLE);
            DISPLAY_SetBlinkingDigits(0);
            setupAlarmTime = false;
            APPLICATION_ChangeModeTo(OM_CLOCK);
        }
        else
        {   switch( menuItem )
            {   case SHOW_ALARMTIME:    // Weckzeit stellen
                    setupAlarmTime = true;
                    DISPLAY_SetBlinkingDigits(0b11000100);
                    SOUND_Start(SOUND_OPEN_MENU);
                    break;
                case ENABLE_ALARM:      // Wecker ein-/ausschalten
                    if( ALARM_IsEnabled() )
                    {   ALARM_Enable(false);
                        SOUND_Start(SOUND_ALARM_DISABLE);
                        DISPLAY_StartAnim(ANIM_ALARM_DISABLE);
                    }
                    else
                    {   ALARM_Enable(true);
                        SOUND_Start(SOUND_ALARM_ENABLE);
                        DISPLAY_StartAnim(ANIM_ALARM_ENABLE);
                    }
                    APPLICATION_ChangeModeTo(OM_CLOCK);
                    break;
                default:                // Menü verlassen
                    SOUND_Start(SOUND_CLOSE_MENU);
                    DISPLAY_StartAnim(ANIM_MENU_CLOSE);
                    APPLICATION_ChangeModeTo(OM_CLOCK);
            }
        }
    }
    SInt08 position = APPLICATION_Encoder();
    if( position != 0 )
    {   if( setupAlarmTime ) ShiftTime(&alarmTime, position);
        else                 ShiftMenuItem(&menuItem, position);
    }
    // Zeitüberschreitung
    if( APPLICATION_UserInteraction() ) TIMER_Restart(&exitTimer);
    if( TIMER_HasExpired(&exitTimer) )
    {   SOUND_Start(SOUND_CLOSE_MENU);
        DISPLAY_StartAnim(ANIM_MENU_CLOSE);
        APPLICATION_ChangeModeTo(OM_CLOCK);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
