﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Einstellungen                                                                       //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2016 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Alarm.h"
#include "Timer.h"
#include "Board.h"
#include "Sound.h"
#include "Storing.h"
#include "String.h"
#include "Display.h"
#include "Settings.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                 // Einstellungen:
{   SI_DISPLAY_BLANK_START,  // · Zeit, bis zur Abschaltung der Anzeige: 1 .. 240 [min]
    SI_MOTION_SENS_DISPLAY,  // · Bewegungsmelder-Empf. bei Anzeige ein: 1 .. 9 [#]
    SI_ALARM_OPERATION,      // · Weckbetriebsart:                       [AlarmOperation]
    SI_ALARM_DURATION,       // · Dauer der Weckalarms:                  1 .. 240 [min]
    SI_ALARM_TYPE_RADIO,     // · true/false = Weckmittel Radio/Signal:  true/false
    SI_SNOOZE_ENABLED,       // · true = Schlummerfunktion aktiv:        true/false
    SI_MOTION_SENS_SNOOZE,   // · Bewegungsmelder-Empf. bei Schlummen:   1 .. 9 [#]
    SI_SNOOZE_DURATION,      // · Dauer der Schlummerphase:              1 .. 99 [min]
    SI_GERMAN_REGION,        // · Deutsches Bundesland:                  [GermanRegion]
    SI_SHOW_HOLIDAY,         // · Anzeige freier Tage                    true/false
    SI_DAYLIGHT_SAVING_MODE, // · Sommerzeiteinstellung:                 [DaylightSavingMode]
    SI_FACTORY_DEFAULTS,     // · Alle Einstellungen zurücksetzen
    SI_RESTART_DEVICE,       // · Gerät neu starten
    SI_EXIT                  // · Einstellungsmenü verlassen
} SettingItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Timer        exitTimer      = TIMER(TIMER_MILLISECONDS, CFG_MODE_SETTINGS_TIMEOUT*1000);
static SettingItem  settingItem    = SI_DISPLAY_BLANK_START;
static SInt16       settingsValue  = INVALID;
static bool         setupAlarmTime = false;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintNumValue(String* string, SInt16 value)
{
	STRING_PrintNum(string, value);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintEnumText(String* string, SInt16 value, const char* text)
{
	String tmpString = STRING;
	STRING_PrintTxtP(&tmpString, text);
	UInt08 len = STRING_Len(&tmpString)+1;
	STRING_PrintTxtP(string, &text[value*len]);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintPrefix(String* string, SettingItem item)
{
	switch( item )
    {   case SI_DISPLAY_BLANK_START:  STRING_PrintLit(string, "LEAUS");  break; // · Zeit, bis zur Abschaltung der Anzeige
        case SI_MOTION_SENS_DISPLAY:  STRING_PrintLit(string, "LEdSEn"); break; // · Bewegungsmelder-Empf. bei Anzeige ein
        case SI_ALARM_OPERATION:      STRING_PrintLit(string, "AO");     break; // · Weckbetriebsart
        case SI_ALARM_DURATION:       STRING_PrintLit(string, "ALEn");   break; // · Dauer der Weckalarms
        case SI_ALARM_TYPE_RADIO:     STRING_PrintLit(string, "At");     break; // · true/false = Weckmittel Radio/Signal
        case SI_SNOOZE_ENABLED:       STRING_PrintLit(string, "Snoo");   break; // · true = Schlummerfunktion aktiv
        case SI_MOTION_SENS_SNOOZE:   STRING_PrintLit(string, "SnoSEn"); break; // · Bewegungsmelder-Empf. bei Schlummen
        case SI_SNOOZE_DURATION:      STRING_PrintLit(string, "SnoLE");  break; // · Dauer der Schlummerphase
        case SI_GERMAN_REGION:        STRING_PrintLit(string, "LAnd");   break; // · Deutsches Bundesland
        case SI_SHOW_HOLIDAY:         STRING_PrintLit(string, "AnFr");   break; // · Anzeige freier Tage
        case SI_DAYLIGHT_SAVING_MODE: STRING_PrintLit(string, "dSt");    break; // · Sommerzeiteinstellung
        case SI_FACTORY_DEFAULTS:     STRING_PrintLit(string, "rESEt");  break; // · Alle Einstellungen zurücksetzen
        case SI_RESTART_DEVICE:       STRING_PrintLit(string, "StArt");  break; // · Gerät neu starten
        default:                      STRING_PrintLit(string, "--");			// · Einstellungsmenü verlassen
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
StartFlashingPrefix(SettingItem item)
{
    String string = STRING;
    PrintPrefix(&string, item);
    UInt08 len = STRING_Len(&string);
    static const UInt08 bitMask[] PROGMEM   = { 0b00000000,
                                                0b10000000,
                                                0b11000000,
                                                0b11100000,
                                                0b11110000,
                                                0b11111000,
                                                0b11111100,
                                                0b11111110,
                                                0b11111111 };
    UInt08 digits = pgm_read_byte(&bitMask[len]);
    DISPLAY_SetBlinkingDigits(digits);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintValue(String* string, SettingItem item, SInt16 value)
{
	static const char alarmOpEnumText[] PROGMEM = "1-t  \0"    // Täglich
											      "2-tE \0"    // Täglich außer an Wochenenden (Sa/So)
											      "3-tEF\0"    // Täglich außer an Wochenenden und Feiertagen
											      "4-E  \0"    // Nur an Wochenenden
											      "5-EF \0"    // Nur an Wochenenden und Feiertagen
											      "6-F  \0"    // Nur an Feiertagen
											      "7-Ein";     // Einmalig (Keine Wiederholung)
	static const char alarmTypeEnumText[] PROGMEM = "Sound\0"  // Weckalarm mit Ton
											        "rAdio";   // Weckalarm mit Radio
	static const char regionEnumText[] PROGMEM = "L01\0"       // Baden-Württemberg
											     "L02\0"       // Bayern
											     "L03\0"       // Berlin
											     "L04\0"       // Brandenburg
											     "L05\0"       // Bremen
											     "L06\0"       // Hamburg
											     "L07\0"       // Hessen
											     "L08\0"       // Mecklenburg-Vorpommern
											     "L09\0"       // Niedersachsen
											     "L10\0"       // Nordrhein-Westfalen
											     "L11\0"       // Rheinland-Pfalz
											     "L12\0"       // Saarland
											     "L13\0"       // Sachsen
											     "L14\0"       // Sachsen-Anhalt
											     "L15\0"       // Schleswig-Holstein
											     "L16";        // Thüringen
	static const char onOffEnumText[] PROGMEM = "AUS\0"        // Allgemein Aus
                                                "Ein";         // Allgemein Ein
	static const char yesNoEnumText[] PROGMEM = "--\0"         // Allgemein Nein
                                                "JA";          // Allgemein Ja
    static const char daylightSavingMode[] PROGMEM = "AUS \0"  // Normalzeit
											         "Ein \0"  // Sommerzeit
											         "Auto\0"; // Automatische Umschaltung

	switch( item )
    {   case SI_DISPLAY_BLANK_START:  PrintNumValue(string, value);                     break;
        case SI_MOTION_SENS_DISPLAY:  PrintNumValue(string, value);                     break;
        case SI_ALARM_OPERATION:      PrintEnumText(string, value, alarmOpEnumText);    break;
        case SI_ALARM_DURATION:       PrintNumValue(string, value);	                    break;
        case SI_ALARM_TYPE_RADIO:     PrintEnumText(string, value, alarmTypeEnumText);  break;
        case SI_SNOOZE_ENABLED:       PrintEnumText(string, value, onOffEnumText);	    break;
        case SI_MOTION_SENS_SNOOZE:   PrintNumValue(string, value);                     break;
        case SI_SNOOZE_DURATION:      PrintNumValue(string, value);	                    break;
        case SI_GERMAN_REGION:        PrintEnumText(string, value, regionEnumText);     break;
        case SI_SHOW_HOLIDAY:         PrintEnumText(string, value, onOffEnumText);	    break;
        case SI_DAYLIGHT_SAVING_MODE: PrintEnumText(string, value, daylightSavingMode); break;
        case SI_FACTORY_DEFAULTS:     PrintEnumText(string, value, yesNoEnumText);      break;
        case SI_RESTART_DEVICE:       PrintEnumText(string, value, yesNoEnumText);	    break;
        default:                      STRING_PrintLit(string, "---");
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
ModifyValue(SInt16 min, SInt16* value, SInt16 max)
{
	if( APPLICATION_Key() ) return false;	// Beendet
    bool boolean = min == (SInt16)false && max == (SInt16)true;
    SInt16 tmpValue = *value;
    SInt08 position = APPLICATION_Encoder();
    if( position != 0 )
    {   if( boolean )
        {   while( position != 0 )
            {   if( tmpValue == min ) tmpValue = max;
                else                  tmpValue = min;
                SOUND_Start(SOUND_ENC_TICK);
                if( position > 0 ) position--;
                else               position++;
            }
        }
        else
        {   while( position > 0 )
            {	if( tmpValue < max )
                {   tmpValue += 1;
				    SOUND_Start(SOUND_ENC_TICK);
                }
                position--;
            }
		    while( position < 0 )
		    {	if( tmpValue > min )
                {   tmpValue -= 1;
				    SOUND_Start(SOUND_ENC_TICK);
                }
                position++;
		    }
        }
	}
    *value = tmpValue;
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
Load(SettingItem item)
{
	UInt08 value;
	switch( item )
	{   case SI_DISPLAY_BLANK_START:  STORING_Load(STO_DisplayBlankStart , &value); break;
		case SI_MOTION_SENS_DISPLAY:  STORING_Load(STO_MotionSensDisplay , &value); break;
        case SI_ALARM_OPERATION:      STORING_Load(STO_AlarmOperation    , &value); break;
        case SI_ALARM_DURATION:       STORING_Load(STO_AlarmDuration     , &value); break;
        case SI_ALARM_TYPE_RADIO:     STORING_Load(STO_AlarmTypeRadio    , &value); break;
        case SI_SNOOZE_ENABLED:       STORING_Load(STO_SnoozeEnabled     , &value); break;
        case SI_MOTION_SENS_SNOOZE:   STORING_Load(STO_MotionSensSnooze  , &value); break;
        case SI_SNOOZE_DURATION:      STORING_Load(STO_SnoozeDuration    , &value); break;
        case SI_GERMAN_REGION:        STORING_Load(STO_GermanRegion      , &value); break;
        case SI_SHOW_HOLIDAY:         STORING_Load(STO_ShowHoliday       , &value); break;
        case SI_DAYLIGHT_SAVING_MODE: STORING_Load(STO_DaylightSavingMode, &value); break;
        case SI_FACTORY_DEFAULTS:     value = 0;
        case SI_RESTART_DEVICE:       value = 0;
        default:                      value = 0;
	}
	return value;
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
Modify(SettingItem item, SInt16* value)
{
	switch( item )
    {   case SI_DISPLAY_BLANK_START: return ModifyValue((SInt16)       1, value, (SInt16)          99);
        case SI_MOTION_SENS_DISPLAY: return ModifyValue((SInt16)       1, value, (SInt16)           9);
        case SI_ALARM_OPERATION:     return ModifyValue((SInt16)AO_DAILY, value, (SInt16)AO_ONCE_ONLY);
        case SI_ALARM_DURATION:      return ModifyValue((SInt16)       1, value, (SInt16)         240);
        case SI_ALARM_TYPE_RADIO:    return ModifyValue((SInt16)   false, value, (SInt16)        true);
        case SI_SNOOZE_ENABLED:      return ModifyValue((SInt16)   false, value, (SInt16)        true);
        case SI_MOTION_SENS_SNOOZE:  return ModifyValue((SInt16)       1, value, (SInt16)           9);
        case SI_SNOOZE_DURATION:     return ModifyValue((SInt16)       1, value, (SInt16)          99);
        case SI_GERMAN_REGION:       return ModifyValue((SInt16)   GR_BW, value, (SInt16)       GR_TH);
        case SI_SHOW_HOLIDAY:        return ModifyValue((SInt16)   false, value, (SInt16)        true);
        case SI_DAYLIGHT_SAVING_MODE: return ModifyValue((SInt16)DSM_OFF, value, (SInt16)    DSM_AUTO);
        case SI_FACTORY_DEFAULTS:    return ModifyValue((SInt16)   false, value, (SInt16)        true);
        case SI_RESTART_DEVICE:      return ModifyValue((SInt16)   false, value, (SInt16)        true);
        default:                     return false;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
Save(SettingItem item, SInt16 value)
{
	UInt08 tmpValue = value;
	switch( item )
	{   case SI_DISPLAY_BLANK_START:  STORING_Save(STO_DisplayBlankStart , &tmpValue); break;
		case SI_MOTION_SENS_DISPLAY:  STORING_Save(STO_MotionSensDisplay , &tmpValue); break;
        case SI_ALARM_OPERATION:      STORING_Save(STO_AlarmOperation    , &tmpValue); break;
        case SI_ALARM_DURATION:       STORING_Save(STO_AlarmDuration     , &tmpValue); break;
        case SI_ALARM_TYPE_RADIO:     STORING_Save(STO_AlarmTypeRadio    , &tmpValue); break;
        case SI_SNOOZE_ENABLED:       STORING_Save(STO_SnoozeEnabled     , &tmpValue); break;
        case SI_MOTION_SENS_SNOOZE:   STORING_Save(STO_MotionSensSnooze  , &tmpValue); break;
        case SI_SNOOZE_DURATION:      STORING_Save(STO_SnoozeDuration    , &tmpValue); break;
        case SI_GERMAN_REGION:        STORING_Save(STO_GermanRegion      , &tmpValue); break;
        case SI_SHOW_HOLIDAY:         STORING_Save(STO_ShowHoliday       , &tmpValue); break;
        case SI_DAYLIGHT_SAVING_MODE: STORING_Save(STO_DaylightSavingMode, &tmpValue); break;
		case SI_FACTORY_DEFAULTS:     if( tmpValue == 1 ) STORING_RestoreDefaults();
        case SI_RESTART_DEVICE:       if( tmpValue == 1 )
                                      {   DateTime invalid = DATETIME_INVALID;
                                          BOARD_SetClock(&invalid);
                                          BOARD_Reset();
                                      }
                                      break;
        default:                      ;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
SETTINGS_Enter(void)
{
    settingItem = SI_DISPLAY_BLANK_START;
    setupAlarmTime = false;
	settingsValue = Load(settingItem);
    DISPLAY_StartAnim(ANIM_SETTINGS_OPEN);
    DISPLAY_SetBlinkingDigits(0);
    SOUND_Start(SOUND_SETTINGS_OPEN);
    TIMER_Restart(&exitTimer);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
SETTINGS_Proceed(void)
{
    // Ausgabe
    String string = STRING;
	PrintPrefix(&string, settingItem);
	STRING_PrintChar(&string, ' ');
	PrintValue(&string,settingItem, settingsValue);
	while( STRING_Len(&string) < 8 ) STRING_PrintChar(&string, ' ');
	DISPLAY_Print(&string);
    // Aktion
    if( setupAlarmTime )
    {   if( !Modify(settingItem, &settingsValue) )
		{	Save(settingItem, settingsValue);
			setupAlarmTime = false;
			SOUND_Start(SOUND_CHANGE_VALUE_END);
            DISPLAY_SetBlinkingDigits(0);
		}
    }
	else
	{	if( APPLICATION_Key() )
		{	if( settingItem == SI_EXIT )
            {   DISPLAY_SetBlinkingDigits(0);
                DISPLAY_StartAnim(ANIM_SETTINGS_CLOSE);
                SOUND_Start(SOUND_SETTINGS_CLOSE);
                APPLICATION_ChangeModeTo(OM_CLOCK);
            }
			else
            {   SOUND_Start(SOUND_CHANGE_VALUE_BEGIN);
                setupAlarmTime = true;
                StartFlashingPrefix(settingItem);
            }
		}
		else
		{	SInt08 position = APPLICATION_Encoder();
			if( position != 0 )
			{   while( position > 0 )
				{  if( settingItem != SI_EXIT )
				   {   settingItem = (SettingItem)((UInt08)settingItem+1);
					   SOUND_Start(SOUND_ENC_TICK);
				   }
				   position--;
				}
				while( position < 0 )
				{	if( settingItem != SI_DISPLAY_BLANK_START )
					{   settingItem  = (SettingItem)((UInt08)settingItem-1);
						SOUND_Start(SOUND_ENC_TICK);
					}
					position++;
				}
				settingsValue = Load(settingItem);
			}
		}
	}
    // Zeitüberschreitung
    if( APPLICATION_UserInteraction() ) TIMER_Restart(&exitTimer);
    if( TIMER_HasExpired(&exitTimer) )
    {   DISPLAY_SetBlinkingDigits(0);
        DISPLAY_StartAnim(ANIM_SETTINGS_CLOSE);
        SOUND_Start(SOUND_SETTINGS_CLOSE);
        APPLICATION_ChangeModeTo(OM_CLOCK);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
